package com.csg.cs.ws.locator;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;

public class SmartSEILocatorFactory {

  /**
   * @param locator SEI Locator interface
   * @return Returns a SEILocator that on every method invocation invokes the method getSEI
   * of given SEILocator before the requested method will be invoked.
   */
  @SuppressWarnings("unchecked")
  public static <SEI> SEILocator<SEI> createLocatingPerSEIMethodInvocationSEILocator(
      final SEILocator<SEI> locator) {

    InvocationHandler handler = new InvocationHandler() {
      public Object invoke(Object proxy, Method method, Object[] args)
          throws Throwable {
        try {
          SEI sei = locator.getSEI();
          return method.invoke(sei, args);
        }
        catch (InvocationTargetException ie) {
          Throwable cause = ie.getCause();
          throw cause == null ? ie : cause;
        }
      }
    };
    Class<?>[] allInterfaces = locator.getSEI().getClass().getInterfaces();
    final SEI proxiedSEI = (SEI)Proxy.newProxyInstance(
        SmartSEILocatorFactory.class.getClassLoader(), allInterfaces,
        handler);
    return new SEILocator<SEI>() {
      public SEI getSEI() {
        return proxiedSEI;
      }
    };
  }
  
  /**
   * Returns a SEILocator that caches (and reuse) the SEI object.
   * @param locator <SEI>
   * @param makeThreadSafe boolean to make thread safe (use thread local)
   * @return Returns a SEILocator
   */
  public static <SEI> SEILocator<SEI> createCachingSEILocator(final SEILocator<SEI> locator, boolean makeThreadSafe) {
    if (makeThreadSafe) {
      return new ThreadLocalSEILocator<SEI>(locator);
    }
    else {
      final SEI sei = locator.getSEI();
        return new SEILocator<SEI>() {
          public SEI getSEI() {
            return sei;
          }
        };
    }
  } 
    
  
}
