package com.csg.cs.ws.locator;

import java.lang.reflect.Constructor;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import javax.xml.ws.WebServiceException;

import com.csg.cs.core.base.config.ConfigException;
import com.csg.cs.core.base.config.ConfigHelper;
import com.csg.cs.core.base.config.Configuration;
import com.csg.cs.core.base.logging.Level;
import com.csg.cs.core.base.logging.Logger;
import com.csg.cs.core.base.logging.LoggerHelper;

public class UriSEILocator<SEI> implements SEILocator<SEI> {

  private static final Logger logger = LoggerHelper
      .getLogger(UriSEILocator.class);

  private static Map<URI, SEILocator<?>> cachedLocators = new ConcurrentHashMap<URI, SEILocator<?>>();

  private static final boolean doCaching = !Boolean
      .getBoolean(UriSEILocator.class.getName() + ".disableCache");

  private static final String SEI_LOCATOR_PARAM = "SEILocator";

  private URI uri;

  public UriSEILocator(String uri) {
    try {
      this.uri = new URI(uri);
    } catch (URISyntaxException e) {
      throw new IllegalArgumentException("Failed to create URI from '" + uri
          + "'", e);
    }
  }

  public UriSEILocator(URI uri) {
    this.uri = uri;
  }

  @SuppressWarnings("unchecked")  
  public SEI getSEI() {
    SEILocator<?> locator = cachedLocators.get(uri);
    if (locator == null) {
      locator = createLocator();
      if (doCaching) {
        cachedLocators.put(uri, locator);
      }
    }
    return (SEI)locator.getSEI();
  }

  @SuppressWarnings( {"unchecked"})
  private SEILocator<?> createLocator() {
    Configuration applConfig = ConfigHelper.getConfigManager()
        .getApplicationConfig();
    Map<String, String> params;
    try {
      params = applConfig.getParameters(uri.getScheme());
    } catch (ConfigException e) {
      e.printStackTrace();
      params = new HashMap<String,String>();
    }
    try {
      Map<String, String> p = applConfig.getParameters(uri.getScheme() + "://" + uri.getPath());
      params.putAll(p);
    } catch (ConfigException e) {
      //do nothing if area doesn't exist.
    }
    try {
      Map<String, String> p = applConfig.getParameters(uri.toString());
      params.putAll(p);
    } catch (ConfigException e) {
      throw new WebServiceException("Failed to read config for '" + uri + "'",
          e);
    }
    params.put("uri", uri.toString());
    if (logger.isLoggable(Level.FINE)) {
      logger.fine("creating SEILocator for '" + uri + "'. Parameters = "
          + params.toString());
    }

    String locatorClassName = params.get(SEI_LOCATOR_PARAM);
    if (locatorClassName == null) {
      throw new WebServiceException("Mandatory parameter " + SEI_LOCATOR_PARAM
          + " not set for '" + uri + "'");
    }
    try {
      Class locatorClass = Class.forName(locatorClassName);
      Constructor c = locatorClass.getConstructor(Map.class);
      SEILocator<SEI> locator = (SEILocator<SEI>)c.newInstance(params);
      String cache = params.get("cache");
      if (cache == null) {
        cache = "threadsafe";
      }
      if (!cache.equalsIgnoreCase("false")) {
        boolean threadSafe = false;
        if (cache.equalsIgnoreCase("threadsafe")) {
          threadSafe = true;
        }
        locator = SmartSEILocatorFactory.createCachingSEILocator(locator,
            threadSafe);
      }
      String locatePerMethodInvocation = params
          .get("locatePerSEIMethodInvocation");
      if (locatePerMethodInvocation == null) {
        locatePerMethodInvocation = "true";
      }
      if (!locatePerMethodInvocation.equalsIgnoreCase("false")) {
        locator = SmartSEILocatorFactory
            .createLocatingPerSEIMethodInvocationSEILocator(locator);
      }
      return locator;
    } catch (Exception e) {
      throw new WebServiceException(
          "Failed to create SEILocator instance for '" + uri + "'", e);
    }
  }
  
}
