package com.csg.cs.ws.locator.j2ee;

import javax.xml.namespace.QName;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.Service;
import javax.xml.ws.WebServiceFeature;

import com.csg.cs.security.wls.net.SSL.wsee.WSEETransportConfigurator;
import com.csg.cs.ws.locator.SEIContextDecoratorFactory;
import com.csg.cs.ws.locator.SEILocator;

/**
 * A {@link SEILocator} to lookup generated JAX-WS service end-point interface
 * (SEI) in a JAP WLS environment. This class encapsulate CS specific code to
 * configure the SEI (WSEETransportConfigurator) and to add WsHeader handler
 * (WsHeaderPropagator) which is responsible for context propagation.
 * 
 * <p>
 * If client code is running in an EJB it's recommended to store SEI simply in
 * an EJB private member, not in a ThreadLocal.
 * 
 * @author A530732
 * 
 * @param <SRV>
 *          Generated JAX-WS service interface
 * @param <SEI>
 *          Generated JAW-WS service end-point interface (port type)
 */
public final class WlsSEILocator<SRV, SEI> implements SEILocator<SEI> {

  /**
   * Generated JAX-WS service interface class.
   */
  private Class<SRV> serviceClass = null;

  /**
   * JAX-WS service interface.
   */
  private Service service = null;

  /**
   * Generated JAX-WS service end-point interface class.
   */
  private Class<SEI> seiClass = null; // 

  /**
   * name of port defined in WSDL.
   */
  private String portName = null;

  /**
   * serviceName of EndpointAddress configured in generated
   * config/datalayer/wsse-config.xml.
   */
  private String serviceNameOfEndpointAddress = null;

  /**
   * Constructor. First 3 paramaeters are elements mapped from WSDL to Java.
   * 
   * @param serviceClass
   *          JAX-WS generated Service class
   * @param serviceEndpointInterfaceClass
   *          JAX-WS generated end-point interface (port type) class
   * @param portName
   *          the name of the port.
   * @param serviceNameOfEndpointAddress
   *          JAP specific service name of EndpointAddress
   */
  public WlsSEILocator(Class<SRV> serviceClass,
      Class<SEI> serviceEndpointInterfaceClass, String portName,
      String serviceNameOfEndpointAddress) {
    this.serviceClass = serviceClass;
    this.seiClass = serviceEndpointInterfaceClass;
    this.portName = portName;
    this.serviceNameOfEndpointAddress = serviceNameOfEndpointAddress;
  }

  public WlsSEILocator(SRV service, Class<SEI> serviceEndpointInterfaceClass,
      String portName, String serviceNameOfEndpointAddress) {
    this.service = (Service)service;
    this.seiClass = serviceEndpointInterfaceClass;
    this.portName = portName;
    this.serviceNameOfEndpointAddress = serviceNameOfEndpointAddress;
  }

  /**
   * Encapsulates CS specific lookup of an SEI.
   */
  @SuppressWarnings( {"deprecation", "unchecked"})
  public SEI getSEI() {
    // See also
    // https://svn01.csintra.net/svnrep/WebLogic_OneBuild/docs/wl1031/tags/v2.0.0_9/WebLogicOneBuild2.0WebServices.doc
    // how to get the service end-point interface (SEI).
    WebServiceFeature[] wsFeatures = WSEETransportConfigurator
        .getWSPolicyFeatureArray(serviceNameOfEndpointAddress);

    Service svc = service;
    if (svc == null) {
      try {
        svc = (Service)serviceClass.newInstance();
      } catch (Exception e) {
        throw new RuntimeException("Failed to create new instance of class "
            + serviceClass.getName(), e);
      }
    }

    QName portQName = new QName(svc.getServiceName().getNamespaceURI(),
        portName);
    SEI port;
    if (wsFeatures == null) {
      // Hint: If lookup won't be done in a generic way, use get<Named>Port
      // instead!
      port = svc.getPort(portQName, seiClass);
    } else {
      // Hint: If lookup won't be done in a generic way, use get<Named>Port
      // instead!
      port = svc.getPort(portQName, seiClass, wsFeatures);
    }
    // Used to configure the SEI(examples are: endpoint URL,
    // SSLSocketFactory)
    WSEETransportConfigurator.configureService(((BindingProvider)port),
        serviceNameOfEndpointAddress);

    // Used to enable WsHeader propagation.
    // Use Interface SEIContext to propagate values.
    return (SEI)SEIContextDecoratorFactory
        .createWsHeaderPropagatorEnabledDecorator((BindingProvider)port);
  }

}
