package com.csg.cs.ws.locator.jse;

import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;
import java.net.UnknownHostException;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Collections;

import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.TrustManagerFactory;
import javax.net.ssl.X509KeyManager;

import com.csg.cs.security.onepki.PKIConfig;
import com.csg.cs.security.onepki.PKIConstants;
import com.csg.cs.security.onepki.PKILoadStoreParameter;
import com.csg.cs.security.onepki.provider.PKIProvider;
import com.csg.cs.security.onepki.ssl.PKIX509KeyManagerFactoryParameters;
import com.csg.cs.security.onepki.ssl.PKIX509TrustManagerFactoryParameters;

/**
 * A utility class used to provide different method to easy use CS_PKI security
 * provider. <br>
 * Use {@link #addProvider()} to add CS_PKI security providers to list of
 * available java security provider first.<br>
 * Use {@link #isProviderAvailable()} to test whether CS_PKI security provider
 * is available. <br>
 * Use one of the methods named createSSLSocketFactoryWith to create a
 * SSLSocketFactory based on CS_PKI.<br>
 * Use {@link #getPrivateKey(String)} to get a PrivateKey from CS_PKI.
 * Use {@link #getCertificate(String)} to get a X509Certificate from CS_PKI.
 * 
 * @author A530732
 * 
 */
public final class CsPkiUtility {

  private static final String SECURITY_PROVIDER_TYPE = "CS_PKI";

  public static boolean isProviderAvailable() {
    boolean isAvailable = true;
    try {
      KeyStore.getInstance(SECURITY_PROVIDER_TYPE);
    } catch (KeyStoreException e) {
      isAvailable = false;
    }
    // if (System.getenv("ONEPKI_CORE_CONFIG_DIR") == null
    // || System.getenv("ONEPKI_CORE_CREDENTIAL_DIR") == null) {
    // isAvailable = false;
    // }
    return isAvailable;
  }

  public static void addProvider() {
    if (!isProviderAvailable()) {
      PKIProvider.addAsProvider();
    }
  }

  public static SSLSocketFactory createSSLSocketFactoryWithMachineCert() {
    return new CsPkiSocketFactory(PKIConstants.CREDENTIAL_ALIAS_MACHINE);
  }

  public static SSLSocketFactory createSSLSocketFactoryWithSPIDCert(String spid) {
    return new CsPkiSocketFactory(PKIConstants.CERT_TYPE_SPID + "_" + spid);
  }

  public static SSLSocketFactory createSSLSocketFactoryWithCredentialAlias(
      String credAlias) {
    return new CsPkiSocketFactory(credAlias);
  }

  private static KeyStore getKeyStore(String credentialAlias)
      throws KeyStoreException, IllegalArgumentException, IOException,
      NoSuchAlgorithmException, CertificateException {
    // Get KeyStore
    KeyStore store = KeyStore.getInstance("CS_PKI");
    // load
    PKILoadStoreParameter loadParameter = new PKILoadStoreParameter(PKIConfig
        .getCoreConfig(), PKILoadStoreParameter.NONE, Collections
        .singletonList(credentialAlias));
    store.load(loadParameter);
    return store;
  }

  public static PrivateKey getPrivateKey(String credentialAlias) {
    try {
      return (PrivateKey)getKeyStore(credentialAlias).getKey(credentialAlias,
          null);
    } catch (Exception e) {
      throw new SecurityException("Failed to get key of alias "
          + credentialAlias, e);
    }
  }

  public static X509Certificate getCertificate(String credentialAlias) {
    try {
      return (X509Certificate)getKeyStore(credentialAlias).getCertificate(
          credentialAlias);
    } catch (Exception e) {
      throw new SecurityException("Failed to get certificate of alias "
          + credentialAlias, e);
    }
  }

  private static class CsPkiSocketFactory extends SSLSocketFactory {
    private SSLSocketFactory delegate = null;

    private String credAlias = null;

    private CsPkiSocketFactory(String credAlias) throws SecurityException {
      this.credAlias = credAlias;
      delegate = getSSLContext(credAlias).getSocketFactory();
    }

    private static SSLContext getSSLContext(String credAlias)
        throws SecurityException {
      if (credAlias == null) {
        credAlias = PKIConstants.CREDENTIAL_ALIAS_MACHINE;
      }

      PKIConfig coreConfig;
      SSLContext context = null;
      try {
        coreConfig = PKIConfig.getCoreConfig();
        KeyManagerFactory kmf;
        kmf = KeyManagerFactory.getInstance(SECURITY_PROVIDER_TYPE);

        PKIX509KeyManagerFactoryParameters spec = new PKIX509KeyManagerFactoryParameters(
            coreConfig, credAlias);
        kmf.init(spec);

        X509KeyManager km = (X509KeyManager)kmf.getKeyManagers()[0];
        String[] aliases = km.getClientAliases("RSA", null);
        if ((aliases == null) || (aliases.length <= 0)) {
          throw new KeyManagementException(
              "No key manager aliases found for credentials identifier '"
                  + credAlias + "'; please check the credentials "
                  + "configuration files.");
        }

        // create trust manager factory
        TrustManagerFactory tkmf = TrustManagerFactory.getInstance("CS_PKI");
        // init the TrustManagerFactory with custom parameters
        PKIX509TrustManagerFactoryParameters tspec = new PKIX509TrustManagerFactoryParameters(
            coreConfig, PKIConstants.TCAL_GROUP_CSINTERNAL,
            PKIConstants.POLICY_SSL_CLIENT_AUTH,
            PKIConstants.POLICY_SSL_SERVER_AUTH);
        tkmf.init(tspec);

        context = SSLContext.getInstance("TLS", /* "SunJSSE" */"CS_PKI");
        context.init(kmf.getKeyManagers(), tkmf.getTrustManagers(), null);
      } catch (Exception e) {
        throw new SecurityException(
            "Failed to create SSLContext using provider "
                + SECURITY_PROVIDER_TYPE + " and alias " + credAlias, e);
      }
      return context;
    }

    @Override
    public Socket createSocket(Socket s, String host, int port,
        boolean autoClose) throws IOException {
      return delegate.createSocket(s, host, port, autoClose);
    }

    @Override
    public String[] getDefaultCipherSuites() {
      return delegate.getDefaultCipherSuites();
    }

    @Override
    public String[] getSupportedCipherSuites() {
      return delegate.getSupportedCipherSuites();
    }

    @Override
    public Socket createSocket(String host, int port) throws IOException,
        UnknownHostException {
      return delegate.createSocket(host, port);
    }

    @Override
    public Socket createSocket(InetAddress address, int port)
        throws IOException {
      return delegate.createSocket(address, port);
    }

    @Override
    public Socket createSocket(String host, int port, InetAddress localHost,
        int localPort) throws IOException, UnknownHostException {
      return delegate.createSocket(host, port, localHost, localPort);
    }

    @Override
    public Socket createSocket(InetAddress address, int port,
        InetAddress localAddress, int localPort) throws IOException {
      return delegate.createSocket(address, port, localAddress, localPort);
    }

    public String toString() {
      return super.toString() + "#" + credAlias;
    }

  }

}
