package com.csg.cs.ws.locator.jse;

import java.io.IOException;
import java.security.Key;
import java.security.KeyException;
import java.security.KeyPair;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.xml.crypto.AlgorithmMethod;
import javax.xml.crypto.KeySelector;
import javax.xml.crypto.KeySelectorException;
import javax.xml.crypto.KeySelectorResult;
import javax.xml.crypto.MarshalException;
import javax.xml.crypto.XMLCryptoContext;
import javax.xml.crypto.XMLStructure;
import javax.xml.crypto.dsig.CanonicalizationMethod;
import javax.xml.crypto.dsig.DigestMethod;
import javax.xml.crypto.dsig.Reference;
import javax.xml.crypto.dsig.SignatureMethod;
import javax.xml.crypto.dsig.SignedInfo;
import javax.xml.crypto.dsig.Transform;
import javax.xml.crypto.dsig.XMLSignature;
import javax.xml.crypto.dsig.XMLSignatureException;
import javax.xml.crypto.dsig.XMLSignatureFactory;
import javax.xml.crypto.dsig.dom.DOMSignContext;
import javax.xml.crypto.dsig.dom.DOMValidateContext;
import javax.xml.crypto.dsig.keyinfo.KeyInfo;
import javax.xml.crypto.dsig.keyinfo.KeyInfoFactory;
import javax.xml.crypto.dsig.keyinfo.KeyValue;
import javax.xml.crypto.dsig.keyinfo.X509Data;
import javax.xml.crypto.dsig.spec.C14NMethodParameterSpec;
import javax.xml.crypto.dsig.spec.ExcC14NParameterSpec;
import javax.xml.crypto.dsig.spec.TransformParameterSpec;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * Utility class with helper methods to easy use openSAML.
 * 
 * @author A530732
 * 
 */
public class JavaxDsigUtility {

  private DocumentBuilder builder = null;

  private XMLSignatureFactory xmlSignatureFactory = null;

  private DocumentBuilder getBuilder() throws ParserConfigurationException {
    if (builder == null) {
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      factory.setNamespaceAware(true);
      builder = factory.newDocumentBuilder();
    }
    return builder;
  }

  private XMLSignatureFactory getXMLSignatureFactory() {
    if (xmlSignatureFactory == null) {
      xmlSignatureFactory = XMLSignatureFactory.getInstance("DOM");
    }
    return xmlSignatureFactory;
  }

  public void sign(Element assertion, String referenceId, PrivateKey key,
      X509Certificate certificate) throws Exception {

    XMLSignatureFactory factory = XMLSignatureFactory.getInstance("DOM");

    KeyPair keyPair = new KeyPair(certificate.getPublicKey(), key);

    KeyInfoFactory kFactory = factory.getKeyInfoFactory();
    KeyInfo keyInfo = kFactory.newKeyInfo(Collections.singletonList(kFactory
        .newX509Data(Collections.singletonList(certificate))));

    List<Transform> transforms = new ArrayList<Transform>();
    transforms.add(factory.newTransform(Transform.ENVELOPED,
        (TransformParameterSpec)null));
    transforms.add(factory.newTransform(CanonicalizationMethod.EXCLUSIVE,
        (ExcC14NParameterSpec)null));

    Reference ref = factory.newReference("#" + referenceId, factory
        .newDigestMethod(DigestMethod.SHA1, null), transforms, null, null);

    SignedInfo signedInfo = factory.newSignedInfo(factory
        .newCanonicalizationMethod(CanonicalizationMethod.EXCLUSIVE,
            (C14NMethodParameterSpec)null), factory.newSignatureMethod(
        SignatureMethod.RSA_SHA1, null), Collections.singletonList(ref));

    XMLSignature signature = factory.newXMLSignature(signedInfo, keyInfo);
    DOMSignContext signContext = new DOMSignContext(keyPair.getPrivate(),
        assertion);
    signature.sign(signContext);
  }

  /**
   * Helper method to read XML file and to extract SAML-Assertion.
   * 
   * @return a SAML Assertion of version 1 or 2.
   */
  public Element readAssertionFromFile(String fileName) throws SAXException,
      IOException, ParserConfigurationException {
    NodeList list = getBuilder().parse(fileName).getElementsByTagNameNS(
        "urn:oasis:names:tc:SAML:1.0:assertion", "Assertion");
    if (list.getLength() < 1) {
      list = getBuilder().parse(fileName).getElementsByTagNameNS(
          "urn:oasis:names:tc:SAML:2.0:assertion", "Assertion");
    }
    if (list.getLength() < 1) {
      throw new SAXException("Assertion not found in file " + fileName);
    } else {
      return (Element)(list.item(0));
    }
  }

  /**
   * Validates a signed Assertion. The assertion has to contain a KeyInfo with a
   * valid certificate! The included Signature will be validated and the digest
   * value of the referenced Assertion will be validated (=reference
   * validation). There is no certificate validation (for example with a trust
   * store)!
   */

  public void validate(Element assertion) throws MarshalException,
      XMLSignatureException {
    // Validate the signature -- i.e. SAML object is pristine:
    NodeList nl = assertion.getElementsByTagNameNS(XMLSignature.XMLNS,
        "Signature");
    if (nl.getLength() == 0)
      throw new IllegalArgumentException("Cannot find Signature element");

    DOMValidateContext context = new DOMValidateContext(
        new KeyValueKeySelector(), nl.item(0));
    context.setIdAttributeNS(assertion, null, "AssertionID");

    XMLSignature signature = getXMLSignatureFactory().unmarshalXMLSignature(
        context);
    if (!signature.validate(context)) {
      throw new XMLSignatureException("Failed to validate signature");
    }
  }

  public static void main(String[] args) {
    JavaxDsigUtility util = new JavaxDsigUtility();
    try {
      for (int i = 0; i < args.length; i++) {
        if ("-validate".equals(args[i]) && i + 1 < args.length) {
          i++;
          String fileName = args[i];
          Element assertion = util.readAssertionFromFile(fileName);
          util.validate(assertion);
          log("Validation of " + fileName + " done");
        }
      }
    } catch (Exception e) {
      log(null, e);
    }
  }

  private static void log(String msg) {
    System.out.println(msg == null ? "" : msg);
  }

  private static void log(String msg, Throwable t) {
    log(msg);
    t.printStackTrace();
  }

  /**
   * KeySelector that can handle KeyValue and X509Data info.
   */
  private static class KeyValueKeySelector extends KeySelector {
    public KeySelectorResult select(KeyInfo keyInfo,
        KeySelector.Purpose purpose, AlgorithmMethod method,
        XMLCryptoContext context) throws KeySelectorException {
      if (keyInfo == null)
        throw new KeySelectorException("Null KeyInfo object!");

      SignatureMethod sm = (SignatureMethod)method;
      List<?> list = keyInfo.getContent();

      for (int i = 0; i < list.size(); i++) {
        XMLStructure xmlStructure = (XMLStructure)list.get(i);
        PublicKey pk = null;
        try {
          if (xmlStructure instanceof KeyValue)
            pk = ((KeyValue)xmlStructure).getPublicKey();
          else if (xmlStructure instanceof X509Data) {
            for (Object data : ((X509Data)xmlStructure).getContent())
              if (data instanceof X509Certificate)
                pk = ((X509Certificate)data).getPublicKey();
          }
        } catch (KeyException ke) {
          throw new KeySelectorException(ke);
        }

        if (algEquals(sm.getAlgorithm(), pk.getAlgorithm()))
          return new SimpleKeySelectorResult(pk);
      }
      throw new KeySelectorException("No KeyValue element found!");
    }

    /**
     * Test that a formal URI expresses the same algorithm as a conventional
     * short name such as "DSA" or "RSA".
     */
    static boolean algEquals(String algURI, String algName) {
      return (algName.equalsIgnoreCase("DSA") && algURI
          .equalsIgnoreCase(SignatureMethod.DSA_SHA1))
          || (algName.equalsIgnoreCase("RSA") && algURI
              .equalsIgnoreCase(SignatureMethod.RSA_SHA1));
    }

  }

  /**
   * Data structure returned by the key selector to the validation context.
   */
  private static class SimpleKeySelectorResult implements KeySelectorResult {
    private PublicKey pk;

    SimpleKeySelectorResult(PublicKey pk) {
      this.pk = pk;
    }

    public Key getKey() {
      return pk;
    }
  }

}
