package com.csg.cs.ws.locator.jse;

import javax.xml.namespace.QName;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.Service;

import com.csg.cs.ws.locator.SEIContextDecoratorFactory;
import com.csg.cs.ws.locator.SEILocator;

/**
 * A {@link SEILocator} to lookup a generated JAX-WS service end-point interface
 * (SEI) in a JSE environment. This class encapsulate CS specific code to
 * configure the SEI:
 * <ul>
 * <li>registers WsHeader handler (WsHeaderPropagator) which is responsible for
 * context propagation</li>
 * <li>optionally sets end-point URL</li>
 * <li>optionally sets SSL socket factory using standard JAVA provider or
 * CS_PKI</li>
 * <li>optionally configures a WS-Security token provider (UNTP, signed or
 * unsigned SAML 1.1)</li>
 * </ul>
 * 
 * @author A530732
 * 
 * @param <SRV>
 *          Generated JAX-WS service interface
 * @param <SEI>
 *          Generated JAW-WS service end-point interface
 */
public final class JseSEILocator<SRV, SEI> implements SEILocator<SEI> {

  /**
   * Generated JAX-WS service interface class.
   */
  private Class<SRV> serviceClass = null;

  /**
   * JAX-WS service interface.
   */
  private Service service = null;

  /**
   * Generated JAX-WS service end-point interface class.
   */
  private Class<SEI> seiClass = null; // 

  /**
   * name of port defined in WSDL.
   */
  private String portName = null;

  /**
   * End-point URL used during lookup to configure the BindingProvider.
   */
  private String endpointAddress = null;

  /**
   * Used to configure transport security (SSL) and message security
   * (WS-Security token).
   */
  private WsSecurityConfigurator wsSecurityConfigurator = new WsSecurityConfigurator();

  public JseSEILocator(Class<SRV> serviceClass,
      Class<SEI> serviceEndpointInterfaceClass, String portName,
      String endpointAddress) {
    this.serviceClass = serviceClass;
    this.seiClass = serviceEndpointInterfaceClass;
    this.portName = portName;
    this.endpointAddress = endpointAddress;
  }

  public JseSEILocator(Class<SRV> serviceClass,
      Class<SEI> serviceEndpointInterfaceClass, String portName) {
    this.serviceClass = serviceClass;
    this.seiClass = serviceEndpointInterfaceClass;
    this.portName = portName;
  }

  public JseSEILocator(SRV service, Class<SEI> serviceEndpointInterfaceClass,
      String portName, String endpointAddress) {
    this.service = (Service)service;
    this.seiClass = serviceEndpointInterfaceClass;
    this.portName = portName;
    this.endpointAddress = endpointAddress;
  }

  public JseSEILocator(SRV service, Class<SEI> serviceEndpointInterfaceClass,
      String portName) {
    this.service = (Service)service;
    this.seiClass = serviceEndpointInterfaceClass;
    this.portName = portName;
  }

  /**
   * Encapsulates CS specific lookup of an SEI.
   */
  @SuppressWarnings( {"deprecation", "unchecked"})
  public SEI getSEI() {
    Service svc = service;
    if (svc == null) {
      try {
        svc = (Service)serviceClass.newInstance();
      } catch (Exception e) {
        throw new RuntimeException("Failed to create new instance of class "
            + serviceClass.getName(), e);
      }
    }

    QName portQName = new QName(svc.getServiceName().getNamespaceURI(),
        portName);
    SEI port = svc.getPort(portQName, seiClass);

    if (endpointAddress != null) {
      ((BindingProvider)port).getRequestContext().put(
          BindingProvider.ENDPOINT_ADDRESS_PROPERTY, endpointAddress);
    }
    wsSecurityConfigurator.configureBindingProvider((BindingProvider)port);

    // Used to enable WsHeader propagation.
    // Use Interface SEIContext to propagate values.
    return (SEI)SEIContextDecoratorFactory
        .createWsHeaderPropagatorEnabledDecorator((BindingProvider)port);
  }

  /**
   * Used to configure transport security (SSL) and message-Security
   * (WS-Security token).
   * 
   * @return
   */
  public WsSecurityConfigurator getWsSecurityConfigurator() {
    return wsSecurityConfigurator;
  }

  /**
   * Sets end point URL used in {@link #getSEI()}.
   */
  public void setEndpointAddress(String endpointAddress) {
    this.endpointAddress = endpointAddress;
  }

}
