package com.csg.cs.ws.locator.jse;

import javax.xml.namespace.QName;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPHeader;
import javax.xml.soap.SOAPHeaderElement;
import javax.xml.soap.SOAPMessage;
import javax.xml.soap.SOAPPart;
import javax.xml.ws.handler.soap.SOAPMessageContext;

import org.joda.time.DateTime;
import org.opensaml.Configuration;
import org.opensaml.common.SAMLVersion;
import org.opensaml.saml1.core.Assertion;
import org.opensaml.saml1.core.Attribute;
import org.opensaml.saml1.core.AttributeStatement;
import org.opensaml.saml1.core.AttributeValue;
import org.opensaml.saml1.core.AuthenticationStatement;
import org.opensaml.saml1.core.Conditions;
import org.opensaml.saml1.core.ConfirmationMethod;
import org.opensaml.saml1.core.NameIdentifier;
import org.opensaml.saml1.core.Subject;
import org.opensaml.saml1.core.SubjectConfirmation;
import org.opensaml.xml.XMLObjectBuilder;
import org.opensaml.xml.schema.XSAny;
import org.w3c.dom.Element;

/**
 * A WS-Security token propagator that creates an unsigned SAML 1.1 assertion in
 * header of a web service request. <br>
 * The implementation is based on openSAML.
 * 
 * @author A530732
 * 
 */
class OpenSamlUnsignedSaml11Propagator implements WsSecurityPropagator {

  private static final String ISSUER = "http://www.credit-suisse.com";

  private static final QName QNAME_WSSE = new QName(
      "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd",
      "Security", "wsse");

  private SamlUtility util = new SamlUtility();

  private PrincipalProvider principalProvider = null;

  @Override
  public void propagate(SOAPMessageContext context) throws Exception {

    SOAPMessage message = context.getMessage();
    SOAPPart part = message.getSOAPPart();
    SOAPEnvelope envelope = part.getEnvelope();
    SOAPHeader header = envelope.getHeader();
    if (header == null) {
      header = envelope.addHeader();
    }

    SOAPHeaderElement securitySoapElement = header.addHeaderElement(QNAME_WSSE);
    securitySoapElement.setMustUnderstand(true);

    Assertion assertion = createAssertion();
    marshallAssertion(assertion, securitySoapElement);
  }

  @Override
  public void setPrincipalProvider(PrincipalProvider principalProvider) {
    this.principalProvider = principalProvider;
  }

  protected SamlUtility getSamlUtility() {
    return util;
  }

  protected void marshallAssertion(Assertion assertion, Element parentElement)
      throws Exception {
    util.addToElement(assertion, parentElement);
  }

  private Assertion createAssertion() throws Exception {
    DateTime now = new DateTime();

    // Conditions of assertion (required in WAS 7.x when assertion is signed)
    Conditions conditions = util.create(Conditions.class,
        Conditions.DEFAULT_ELEMENT_NAME);
    conditions.setNotBefore(now);
    conditions.setNotOnOrAfter(now.plusYears(100));
    
    // AuthenticationStatement
    AuthenticationStatement authenticationStatement = util.create(
        AuthenticationStatement.class,
        AuthenticationStatement.DEFAULT_ELEMENT_NAME);
    authenticationStatement.setSubject(createSubject());
    authenticationStatement.setAuthenticationInstant(now);
    authenticationStatement
        .setAuthenticationMethod("urn:oasis:names:tc:SAML:1.0:am:unspecified");

    // AttributeStatement
    AttributeStatement attributeStatement = util.create(
        AttributeStatement.class, AttributeStatement.DEFAULT_ELEMENT_NAME);
    attributeStatement.setSubject(createSubject());
    attributeStatement.getAttributes().add(
        createAttribute("Groups", "urn:bea:security:saml:groups", "CS_GROUP"));

    // Assertion
    Assertion assertion = util.create(Assertion.class,
        Assertion.DEFAULT_ELEMENT_NAME);
    assertion.setIssueInstant(now);
    assertion.setIssuer(ISSUER);
    assertion.setVersion(SAMLVersion.VERSION_11);
    assertion.setID(util.generatID());
    assertion.setConditions(conditions);
    assertion.getAuthenticationStatements().add(authenticationStatement);
    assertion.getAttributeStatements().add(attributeStatement);
    return assertion;
  }

  private Subject createSubject() throws Exception {
    NameIdentifier nameIdentifier = util.create(NameIdentifier.class,
        NameIdentifier.DEFAULT_ELEMENT_NAME);
    nameIdentifier.setFormat(NameIdentifier.UNSPECIFIED);
    nameIdentifier.setNameIdentifier(principalProvider.getPrincipal());
    ConfirmationMethod confirmationMethod = util.create(
        ConfirmationMethod.class, ConfirmationMethod.DEFAULT_ELEMENT_NAME);
    confirmationMethod
        .setConfirmationMethod("urn:oasis:names:tc:SAML:1.0:cm:sender-vouches");
    SubjectConfirmation subjectConfirmation = util.create(
        SubjectConfirmation.class, SubjectConfirmation.DEFAULT_ELEMENT_NAME);
    subjectConfirmation.getConfirmationMethods().add(confirmationMethod);
    Subject subject = util.create(Subject.class, Subject.DEFAULT_ELEMENT_NAME);
    subject.setNameIdentifier(nameIdentifier);
    subject.setSubjectConfirmation(subjectConfirmation);
    return subject;
  }

  @SuppressWarnings("unchecked")
  private Attribute createAttribute(String name, String nameSpace, String value) {
    Attribute attribute = util.create(Attribute.class,
        Attribute.DEFAULT_ELEMENT_NAME);
    attribute.setAttributeName(name);
    attribute.setAttributeNamespace(nameSpace);

    // Build attribute values as XMLObjects;
    // there is an AttributeValue interface, but it's apparently dead code
    XMLObjectBuilder builder = Configuration.getBuilderFactory().getBuilder(
        XSAny.TYPE_NAME);

    XSAny attributeValue = (XSAny)builder
        .buildObject(AttributeValue.DEFAULT_ELEMENT_NAME);
    attributeValue.setTextContent(value);

    attribute.getAttributeValues().add(attributeValue);
    return attribute;
  }

}
