package com.csg.cs.ws.locator.jse;

import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.security.GeneralSecurityException;
import java.security.Key;
import java.security.KeyStore;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Enumeration;
import java.util.List;

import javax.xml.crypto.dsig.CanonicalizationMethod;
import javax.xml.crypto.dsig.SignatureMethod;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.opensaml.Configuration;
import org.opensaml.DefaultBootstrap;
import org.opensaml.common.impl.SecureRandomIdentifierGenerator;
import org.opensaml.xml.XMLObject;
import org.opensaml.xml.XMLObjectBuilder;
import org.opensaml.xml.XMLObjectBuilderFactory;
import org.opensaml.xml.io.Marshaller;
import org.opensaml.xml.io.MarshallingException;
import org.opensaml.xml.io.UnmarshallingException;
import org.opensaml.xml.security.keyinfo.KeyInfoHelper;
import org.opensaml.xml.security.x509.BasicX509Credential;
import org.opensaml.xml.signature.KeyInfo;
import org.opensaml.xml.signature.SignableXMLObject;
import org.opensaml.xml.signature.Signature;
import org.opensaml.xml.signature.SignatureValidator;
import org.opensaml.xml.signature.Signer;
import org.opensaml.xml.signature.impl.KeyInfoBuilder;
import org.opensaml.xml.signature.impl.SignatureBuilder;
import org.opensaml.xml.validation.ValidationException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * Utility class with helper methods to easy use openSAML.
 * 
 * @author A530732
 * 
 */
public class SamlUtility {
  private SecureRandomIdentifierGenerator generator = null;

  private DocumentBuilder builder = null;;

  static {
    try {
      DefaultBootstrap.bootstrap();
    } catch (Exception e) {
      throw new SecurityException("SAML bootstraping failed", e);
    }
  }

  private SecureRandomIdentifierGenerator getGenerator()
      throws NoSuchAlgorithmException {
    if (generator == null) {
      generator = new SecureRandomIdentifierGenerator();
    }
    return generator;
  }

  private DocumentBuilder getBuilder() throws ParserConfigurationException {
    if (builder == null) {
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      factory.setNamespaceAware(true);
      builder = factory.newDocumentBuilder();
    }
    return builder;
  }

  public String generatID() throws NoSuchAlgorithmException {
    return getGenerator().generateIdentifier();
  }

  @SuppressWarnings("unchecked")
  public <T> T create(Class<T> cls, QName qname) {
    XMLObjectBuilderFactory factory = Configuration.getBuilderFactory();
    XMLObjectBuilder builder = factory.getBuilder(qname);
    return (T)builder.buildObject(qname);
  }

  /**
   * Helper method to add an XMLObject as a child of a DOM Element.
   */
  public Element addToElement(XMLObject object, Element parent)
      throws IOException, MarshallingException, TransformerException {
    Marshaller out = Configuration.getMarshallerFactory().getMarshaller(object);
    return out.marshall(object, parent);
  }

  /**
   * Helper method to read an XML object from a DOM element.
   */
  public XMLObject fromElement(Element element) throws IOException,
      UnmarshallingException, SAXException {
    return Configuration.getUnmarshallerFactory().getUnmarshaller(element)
        .unmarshall(element);
  }

  /**
   * Helper method to read an XML object from a file.
   */
  public XMLObject readFromFile(String fileName) throws IOException,
      UnmarshallingException, SAXException, ParserConfigurationException {
    return fromElement(getBuilder().parse(fileName).getDocumentElement());
  }

  /**
   * Helper method to read XML file and to extract SAML-Assertion.
   * 
   * @return a SAML Assertion of version 1 or 2.
   */
  public SignableXMLObject readAssertionFromFile(String fileName)
      throws SAXException, IOException, ParserConfigurationException,
      UnmarshallingException {
    NodeList list = getBuilder().parse(fileName).getElementsByTagNameNS(
        "urn:oasis:names:tc:SAML:1.0:assertion", "Assertion");
    if (list.getLength() < 1) {
      list = getBuilder().parse(fileName).getElementsByTagNameNS(
          "urn:oasis:names:tc:SAML:2.0:assertion", "Assertion");
    }
    if (list.getLength() < 1) {
      throw new UnmarshallingException("Assertion not found in file "
          + fileName);
    } else {
      return (SignableXMLObject)fromElement((Element)(list.item(0)));
    }
  }

  /**
   * Helper method to get an XMLObject as a DOM Document.
   */
  public Document asDOMDocument(XMLObject object) throws IOException,
      MarshallingException, TransformerException, ParserConfigurationException {
    Document document = getBuilder().newDocument();
    Marshaller out = Configuration.getMarshallerFactory().getMarshaller(object);
    out.marshall(object, document);
    return document;
  }

  /**
   * Creates a signature for signableObj and marshalls signableObj with included
   * signature into parentElement.
   * 
   * @return if parentElement is null a new Document will be returned else null
   *         will be returned.
   */
  public Document signAndMarshall(SignableXMLObject signableObj,
      Element parentElement, PrivateKey key, X509Certificate certificate)
      throws Exception {

    SignatureBuilder signaturebuilder = (SignatureBuilder)Configuration
        .getBuilderFactory().getBuilder(Signature.DEFAULT_ELEMENT_NAME);
    org.opensaml.xml.signature.Signature signature = signaturebuilder
        .buildObject();
    signature.setSignatureAlgorithm(SignatureMethod.RSA_SHA1);
    signature.setCanonicalizationAlgorithm(CanonicalizationMethod.EXCLUSIVE);
    BasicX509Credential basicx509credential = new BasicX509Credential();
    basicx509credential.setPrivateKey(key);
    signature.setSigningCredential(basicx509credential);

    XMLObjectBuilderFactory xmlobjectbuilderfactory = Configuration
        .getBuilderFactory();
    KeyInfoBuilder keyinfobuilder = (KeyInfoBuilder)xmlobjectbuilderfactory
        .getBuilder(KeyInfo.DEFAULT_ELEMENT_NAME);
    KeyInfo keyinfo = keyinfobuilder.buildObject();
    KeyInfoHelper.addCertificate(keyinfo, certificate);

    signature.setKeyInfo(keyinfo);

    signableObj.setSignature(signature);
    Document doc = null;
    if (parentElement == null) {
      doc = asDOMDocument(signableObj);
    } else {
      addToElement(signableObj, parentElement);
    }

    Signer.signObject(signature);
    return doc;
  }

  /**
   * Validates the signature of obj. Obj has to contain a KeyInfo with a valid
   * certificate! The included Signature will be validated and the digest value
   * of the referenced Assertion will be validated (=reference validation).
   * There is no certificate validation (for example with a trust store)!
   */
  public void validate(SignableXMLObject obj) throws CertificateException,
      ValidationException {
    List<X509Certificate> certs = KeyInfoHelper.getCertificates(obj
        .getSignature().getKeyInfo());
    for (X509Certificate cert : certs) {
      BasicX509Credential cred = new BasicX509Credential();
      cred.setPublicKey(cert.getPublicKey());
      SignatureValidator validator = new SignatureValidator(cred);
      try {
        validator.validate(obj.getSignature());
        return;
      } catch (ValidationException e) {
        if (certs.size() == 1) {
          throw e;
        }
      }
    }
    throw new ValidationException(
        "Signature doesn't contain a certificate that validates");
  }

  /**
   * Get a KeyStore object given the keystore filename and password.
   */
  private static KeyStore getKeyStore(String type, String filename,
      String password) throws GeneralSecurityException, IOException {
    KeyStore store = KeyStore.getInstance(type == null ? KeyStore
        .getDefaultType() : type);
    FileInputStream in = new FileInputStream(filename);
    store.load(in, password == null ? null : password.toCharArray());
    in.close();
    return store;
  }

  private static void transform(Document doc, OutputStream out)
      throws TransformerException {
    TransformerFactory.newInstance().newTransformer().transform(
        new DOMSource(doc), new StreamResult(out));
  }

  private static String toArgString(String s) {
    if (s == null || s.equals("null")) {
      return null;
    } else if (s.equals("empty")) {
      return "";
    }
    return s;
  }

  private static void usage() {
    log("Validates or signs a SAML assertion");
    log("Usage for validating: SamlUtility -validate assertionFile.");
    log("Usage for signing: SamlUtility -sign assertionFile keyStoreType keyStoreFile keyStorePwd keyAlias keyPwd certAlias.");
  }

  /**
   * Validates or signs a SAML assertion.</br> Usage for validating:
   * SamlUtility -validate assertionFile.</br> Usage for signing: SamlUtility
   * -sign assertionFile keyStoreType keyStoreFile keyStorePwd keyAlias keyPwd
   * certAlias.</br>
   * 
   * @param args
   */
  public static void main(String[] args) {
    SamlUtility util = new SamlUtility();
    try {
      for (int i = 0; i < args.length; i++) {
        if ("-validate".equals(args[i]) && i + 1 < args.length) {
          i++;
          String fileName = args[i];
          SignableXMLObject assertion = util.readAssertionFromFile(fileName);
          util.validate(assertion);
          log("Validation of " + fileName + " done");
        } else if ("-sign".equals(args[i]) && i + 7 < args.length) {
          i++;
          String assertionFile = toArgString(args[i]);
          i++;
          String keyStoreType = toArgString(args[i]);
          i++;
          String keyStoreFile = toArgString(args[i]);
          i++;
          String keyStorePwd = toArgString(args[i]);
          i++;
          String keyAlias = toArgString(args[i]);
          i++;
          String keyPwd = toArgString(args[i]);
          i++;
          String certAlias = toArgString(args[i]);
          KeyStore store = getKeyStore(keyStoreType, keyStoreFile, keyStorePwd);
          SignableXMLObject assertion = util
              .readAssertionFromFile(assertionFile);
          Key key = store.getKey(keyAlias, keyPwd == null ? null : keyPwd
              .toCharArray());
          Enumeration<String> aliases = store.aliases();
          while (aliases.hasMoreElements()) {
            log("alias=<" + aliases.nextElement() + ">");
          }
          log("Key=" + key);
          Certificate cert = store.getCertificate(certAlias);
          log("Certificate=" + cert);
          log("--------------------------------------------------------------");
          Document doc = util.signAndMarshall(assertion, null, (PrivateKey)key,
              (X509Certificate)cert);
          OutputStream out = new ByteArrayOutputStream();
          transform(doc, out);
          String signedAssertion = out.toString();
          out.close();
          String signedAssertionFile = "signed_" + assertionFile;
          PrintStream pOut = new PrintStream(new FileOutputStream(
              signedAssertionFile));
          pOut.print(signedAssertion);
          out.close();
          log(signedAssertion);
          log("");
          log("--------------------------------------------------------------");
          log("Signing done. Signed assertion written to "
              + signedAssertionFile);
        } else {
          usage();
          return;
        }
      }
    } catch (Exception e) {
      log(null, e);
    }
  }

  private static void log(String msg) {
    System.out.println(msg == null ? "" : msg);
  }

  private static void log(String msg, Throwable t) {
    log(msg);
    t.printStackTrace();
  }

}
