package com.csg.cs.ws.locator.jse;

import java.security.PrivateKey;
import java.security.cert.X509Certificate;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.SSLSession;
import javax.net.ssl.SSLSocketFactory;
import javax.xml.namespace.QName;
import javax.xml.ws.Binding;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.WebServiceException;
import javax.xml.ws.handler.Handler;
import javax.xml.ws.handler.MessageContext;
import javax.xml.ws.handler.soap.SOAPHandler;
import javax.xml.ws.handler.soap.SOAPMessageContext;

import com.csg.cs.security.onepki.PKIConstants;

/**
 * Used to configure the security relevant features SSLSocketFactory (used for
 * the transport), <br>
 * {@link PrincipalProvider}, TokenType, signing credentials and
 * {@link WsSecurityPropagator} (used for message security).
 * <p>
 * See
 * {@link #setDefaultWsSecurityPropagator(com.csg.cs.ws.util.consumer.jse.WsSecurityConfigurator.TokenType, Class)}
 * how to set a WsSecurityPropagator through a system property.
 * 
 * @author A530732
 * 
 */
public final class WsSecurityConfigurator {

  /**
   * List of supported WSSE token types.
   */
  public static enum TokenType {
    NO_TOKEN, UNTP, SAML11_UNSIGNED, SAML11_SIGNED
  }

  private static final ThreadLocalPrincipalProvider THREAD_LOCAL_PRINCIPAL_PROVIDER = new ThreadLocalPrincipalProvider();

  private static Map<TokenType, Class<?>> propagatorClasses = new HashMap<TokenType, Class<?>>();

  private static TokenType defaultTokenType = null;

  private static PrincipalProvider defaultPrincipalProvider = null;

  private static PrivateKey defaultSigningKey = null;

  private static X509Certificate defaultSigningCertificate = null;

  static {
    setDefaultTokenType(null);
    propagatorClasses.put(TokenType.NO_TOKEN, NoTokenPropagator.class);
    propagatorClasses.put(TokenType.UNTP, UNTPPropagator.class);
    propagatorClasses.put(TokenType.SAML11_UNSIGNED,
        OpenSamlUnsignedSaml11Propagator.class);
    propagatorClasses.put(TokenType.SAML11_SIGNED,
        OpenSamlSignedSaml11Propagator.class);
    String propagators = System.getProperty("WsSecurityPropagators");
    if (propagators != null) {
      StringTokenizer tokens = new StringTokenizer(propagators, "=;,");
      while (tokens.hasMoreTokens()) {
        try {
          TokenType type = TokenType.valueOf(tokens.nextToken());
          Class<?> propagatorCl = Class.forName(tokens.nextToken());
          propagatorClasses.put(type, propagatorCl);
        } catch (ClassNotFoundException e) {
          throw new SecurityException(
              "Failed to load configured WsSecurityPropagator", e);
        }
      }
    }
  }

  WsSecurityConfigurator() {
  }

  private TokenType tokenType = defaultTokenType;

  private WsSecurityPropagator propagator = null;

  private PrincipalProvider principalProvider = createPrincipalProvider();

  private PrivateKey signingKey = defaultSigningKey;

  private X509Certificate signingCertificate = defaultSigningCertificate;

  private SSLSocketFactory socketFactory = createSSLSocketFactory();

  /**
   * SOAPHandler that is responsible to introduce WSSE token into SOAP-Header.
   */
  private static final class WsSecurityHandler implements
      SOAPHandler<SOAPMessageContext> {

    private WsSecurityPropagator propagator = null;

    private WsSecurityHandler(WsSecurityPropagator propagator) {
      this.propagator = propagator;
    }

    @Override
    public Set<QName> getHeaders() {
      return Collections.emptySet();
    }

    @Override
    public void close(MessageContext context) {
    }

    @Override
    public boolean handleFault(SOAPMessageContext context) {
      return true;
    }

    @Override
    public boolean handleMessage(SOAPMessageContext context) {
      boolean isOutbound = ((Boolean)context
          .get(MessageContext.MESSAGE_OUTBOUND_PROPERTY)).booleanValue();
      if (isOutbound) {
        try {
          propagator.propagate(context);
        } catch (Exception e) {
          throw new WebServiceException(
              "Failed to propagate WS security token", e);
        }
      }
      return true;
    }
  }

  /**
   * Sets the {@link #WsSecurityConfigurator.TokenType} used as default if a new
   * instance will be created. The default is TokenType.SAML11_UNSIGNED
   */
  public static void setDefaultTokenType(TokenType defaultType) {
    defaultTokenType = defaultType;
    if (defaultTokenType == null) {
      defaultTokenType = TokenType.SAML11_UNSIGNED;
    }
  }

  /**
   * Associates a {@link WsSecurityConfigurator.TokenType} with a
   * {@link WsSecurityPropagator}. A configurative way to change this
   * association is through system property WsSecurityPropagators. An example:
   * -DWsSecurityPropagators=SAML11_SIGNED,com.csg.cs.ws.util.consumer.jse.OpenSamlSignedSaml11Propagator2
   */
  public static void setDefaultWsSecurityPropagator(TokenType tokenType,
      Class<?> wsSecurityPropagatorCl) {
    propagatorClasses.put(tokenType, wsSecurityPropagatorCl);
  }

  /**
   * Sets the credentials used for signing the WSSE token. system default: If a
   * {@link SigningWsSecurityPropagator} is used and the CS_PKI is available,
   * the credentials from the MACHINE certificate are used. No system defaults
   * in all other cases.
   * 
   * @param defaultKey
   *          the private key.
   * @param defaultCertificate
   *          the certificate that contains the public key.
   */
  public static void setDefaultSigningCredentials(PrivateKey defaultKey,
      X509Certificate defaultCertificate) {
    defaultSigningKey = defaultKey;
    defaultSigningCertificate = defaultCertificate;
  }

  /**
   * The {@link PrincipalProvider} that should be used as default. The system
   * default is {@link CfwPrincipalProvider} if CFW is available, else a
   * {@link ThreadLocalPrincipalProvider}.
   */
  public static void setDefaultPrincipalProvider(
      PrincipalProvider principalProvider) {
    defaultPrincipalProvider = principalProvider;
  }

  private static PrincipalProvider createPrincipalProvider() {
    if (defaultPrincipalProvider != null) {
      return defaultPrincipalProvider;
    }
    if (CfwPrincipalProvider.isAvailable()) {
      return new CfwPrincipalProvider();
    } else {
      return THREAD_LOCAL_PRINCIPAL_PROVIDER;
    }
  }

  private static SSLSocketFactory createSSLSocketFactory() {
    if (CsPkiUtility.isProviderAvailable()) {
      return CsPkiUtility.createSSLSocketFactoryWithMachineCert();
    }
    return null;
  }

  private static PrivateKey getPrivateKey() {
    if (CsPkiUtility.isProviderAvailable()) {
      return CsPkiUtility.getPrivateKey(PKIConstants.CREDENTIAL_ALIAS_MACHINE);
    }
    return null;
  }

  private static X509Certificate getCertificate() {
    if (CsPkiUtility.isProviderAvailable()) {
      return CsPkiUtility.getCertificate(PKIConstants.CREDENTIAL_ALIAS_MACHINE);
    }
    return null;
  }

  /**
   * Sets the tokenType that has to be used. A TokenType is associated with a
   * {@link WsSecurityPropagator}. If this method will not be called default
   * TokenType will be used.
   */
  public void setTokenType(TokenType tokenType) {
    this.tokenType = tokenType;
  }

  /**
   * Returns the TokenType currently in use. A value of null is identically to
   * the default TokenType.
   */
  public TokenType getTokenType() {
    return tokenType;
  }

  /**
   * Sets the WsSecurityPropagator to be used. If this method is used, the
   * TokenType is irrelevant. Using this method allows to use a propagator of
   * choice per WsSecurityConfigurator instance.
   */
  public void setWsSecurityPropagator(WsSecurityPropagator propagator) {
    this.propagator = propagator;
  }

  /**
   * Sets the credentials used to sign the WSSE token. If not set through this
   * method the default signing credentials will be used.
   */
  public void setSigningCredentials(PrivateKey key, X509Certificate certificate) {
    this.signingKey = key;
    this.signingCertificate = certificate;
  }

  /**
   * Sets the {@link PrincipalProvider} to be used. If not set through this
   * method the default PrincipalProvider will be used.
   */
  public void setPrincipalProvider(PrincipalProvider principalProvider) {
    this.principalProvider = principalProvider;
  }

  public PrincipalProvider getPrincipalProvider() {
    return principalProvider;
  }

  /**
   * Sets the SSLSocketFactory to be used to establish connection.
   * 
   * @param socketFactory
   */
  public void setSSLSocketFactory(SSLSocketFactory socketFactory) {
    this.socketFactory = socketFactory;
  }

  public SSLSocketFactory getSSLSocketFactory() {
    return socketFactory;
  }

  @SuppressWarnings("unchecked")
  void configureBindingProvider(BindingProvider bindingProvider)
      throws WebServiceException {
    Binding binding = bindingProvider.getBinding();
    List<Handler> handlers = binding.getHandlerChain();
    for (Handler h : handlers) {
      if (h instanceof WsSecurityPropagator) {
        throw new WebServiceException("SOAPHandler of type "
            + WsSecurityPropagator.class.getName() + " already registered!");
      }
    }
    try {
      if (propagator == null) {
        Class propagatorCl = propagatorClasses.get(tokenType);
        propagator = (WsSecurityPropagator)propagatorCl.newInstance();
      }
      propagator.setPrincipalProvider(principalProvider);
      if (propagator instanceof SigningWsSecurityPropagator) {
        if (signingKey == null || signingCertificate == null) {
          // try to get default values
          signingKey = getPrivateKey();
          signingCertificate = getCertificate();
        }
        ((SigningWsSecurityPropagator)propagator).setSigningCredentials(
            signingKey, signingCertificate);
      }
    } catch (Exception e) {
      throw new WebServiceException(
          "Failed to initialize WsSecurityPropagator ", e);
    }
    SOAPHandler<SOAPMessageContext> handler = new WsSecurityHandler(propagator);

    // hostname verifier used in cases where no wls is involved (ex. standalone
    // consumer based on jdks wsimport)
    HostnameVerifier hv = new HostnameVerifier() {
      public boolean verify(String arg0, SSLSession arg1) {
        return true;
      }
    };
    bindingProvider.getRequestContext().put(
        // com.sun.xml.internal.ws.client.BindingProviderProperties.HOSTNAME_VERIFIER
            "com.sun.xml.internal.ws.transport.https.client.hostname.verifier", hv);
    bindingProvider.getRequestContext().put(
            "com.sun.xml.ws.transport.https.client.hostname.verifier", hv);
    if (socketFactory != null) {
      // used where wls clientgen is used and and wls in classpath
      bindingProvider.getRequestContext().put(
          "com.sun.xml.ws.transport.https.client.SSLSocketFactory",
          socketFactory);
      // used when jdk wsimport is used.
      bindingProvider.getRequestContext().put(
          // com.sun.xml.internal.ws.client.BindingProviderProperties.SSL_SOCKET_FACTORY
          "com.sun.xml.internal.ws.transport.https.client.SSLSocketFactory",
          socketFactory);
    }
    handlers.add(handler);
    binding.setHandlerChain(handlers);
  }

  public String toString() {
    String p = propagator == null ? propagatorClasses.get(tokenType).getName()
        : propagator.toString();
    return "[" + getClass().getName() + ": tokenType=" + tokenType
        + ", WsSecurityPropagator=" + p + ", principalProvider="
        + principalProvider + ", socketFactory=" + socketFactory + "]";
  }

}
