package com.csg.cs.services.samples.consumer;

import java.io.PrintWriter;
import java.util.Map;
import java.util.UUID;

import javax.xml.ws.BindingProvider;
import javax.xml.ws.soap.SOAPFaultException;

import com.csg.cs.core.base.logging.Logger;
import com.csg.cs.core.base.logging.LoggerHelper;
import com.csg.cs.corex.ctxtransmitter.ws.consumer.WsHeaderPropagator;
import com.csg.cs.schemas.basetypes.tech_1.ExTraceEntry1;
import com.csg.cs.schemas.basetypes.tech_1.LogEntry1;
import com.csg.cs.schemas.tech.intsw.intserv.PCustomer1;
import com.csg.cs.services.inquiry.ping_1_0.PingRequest;
import com.csg.cs.services.inquiry.ping_1_0.PingResponse;
import com.csg.cs.services.tech_customersample_1.TECHCustomerSample1PortType;
import com.csg.cs.services.tech_customersample_2.ServiceExceptionFaultMessage;
import com.csg.cs.services.tech_customersample_2.TECHCustomerSample2PortType;
import com.csg.cs.services.tech_customersample_2.createcustomer_2_0.CreateCustomer20;
import com.csg.cs.services.tech_customersample_2.createcustomer_2_0.CreateCustomer20Response;
import com.csg.cs.ws.locator.SEIContext;
import com.csg.cs.ws.locator.SEILocator;

/**
 * Handles a request to a TECH_CustomerSample_1_PortType or _2_PortType.<br>
 * The class is suitable to use in a J2EE servlet or in a JSE environment.<br>
 * Method {@link #createInstance(PrintWriter, Map, SEILocator, SEILocator)} acts
 * as a factory.
 * 
 * @author A530732
 */
public class CustomerSampleRequestHandler {

  private static final Logger logger = LoggerHelper
      .getLogger(CustomerSampleRequestHandler.class);

  /**
   * Used to write any output during request handling.
   */
  private PrintWriter writer = null;

  /**
   * The PortType used to invoke the business operation of the web service.
   */
  private TECHCustomerSample2PortType sei = null;

  /**
   * The request parameters.
   */
  private Map<String, String[]> parameters = null;

  /**
   * Factory method to create an instance.
   * 
   * @param writer
   *          Used to write any output during request handling
   * @param parameters
   *          the request parameters. Parameter interfaceGroup is used within
   *          create method to decide which web service should be called. See
   *          also documentation of {@link #handleRequest()} to see parameters
   *          used during request handling.
   * @param sample1Loc
   *          the SEI of version 1 of CustomerSample port type
   * @param sample2Loc
   *          the SEI of version 2 of CustomerSample port type
   * @return an instance ready to handle a request.
   */
  public static CustomerSampleRequestHandler createInstance(PrintWriter writer,
      Map<String, String[]> parameters,
      SEILocator<TECHCustomerSample1PortType> sample1Loc,
      SEILocator<TECHCustomerSample2PortType> sample2Loc) {

    String interfaceGroup = getParameter(parameters, "interfaceGroup");
    if (interfaceGroup == null || interfaceGroup.trim().length() == 0) {
      interfaceGroup = "TECH_CustomerSample_1";
    }

    CustomerSampleRequestHandler customerSampleRequestHandler = new CustomerSampleRequestHandler();
    if (interfaceGroup.startsWith("TECH_CustomerSample_1")) {
      customerSampleRequestHandler
          .setSEI(new TECHCustomerSample1PortTypeAdapter(sample1Loc.getSEI()));
    } else {
      customerSampleRequestHandler.setSEI(sample2Loc.getSEI());
    }
    customerSampleRequestHandler.setPrintWriter(writer);
    customerSampleRequestHandler.setParameterMap(parameters);
    return customerSampleRequestHandler;
  }

  private CustomerSampleRequestHandler() {
  }

  private void setSEI(TECHCustomerSample2PortType sei) {
    this.sei = sei;
  }

  private void setPrintWriter(PrintWriter writer) {
    this.writer = writer;
  }

  private void setParameterMap(Map<String, String[]> parameters) {
    this.parameters = parameters;
  }

  private static String getParameter(Map<String, String[]> params, String key) {
    String v = null;
    String[] vArray = params.get(key);
    if (vArray != null && vArray.length > 0) {
      v = vArray[0];
    }
    return v;
  }

  /**
   * Gets a parameter from preseted parameter list. Returns null if not found.
   */
  private String getParameter(String key) {
    return getParameter(parameters, key);
  }

  /**
   * Gets a parameter from preseted parameter list. Returns defaultValue if not
   * found.
   */
  private String getParameter(String key, String defaultValue) {
    String s = getParameter(key);
    if (s == null || "".equals(s)) {
      s = defaultValue;
    }
    return s;
  }

  /**
   * Used parameters during request handling: interfaceGroup, interface,
   * targetBUId, customerLastName*, customerFirstName*. *only if interface
   * createCustomer is invoked.
   */
  public void handleRequest() {
    String interfaceGroup = getParameter("interfaceGroup");
    String targetBUId = getParameter("targetBUId", "");
    String interfaceName = getParameter("interface", "ping_1_0");
    String requestId = UUID.randomUUID().toString();
    String origBuid = "0011";

    writer.println("<h1>Test for " + sei.toString() + "</h1>");
    writer.println("<b>interfaceGroup:</b> " + interfaceGroup + "<br/>");
    writer.println("<b>targetBUId:</b> " + targetBUId + "<br/>");
    writer.println("<b>interface:</b> " + interfaceName + "<br/>");
    writer.println("<b>requestId:</b> " + requestId + "<br/>");

    // Used if for each call another targetBUId has to be propagate.
    // If always the same has to be propagated, the propagated value
    // can be configured!
    if (targetBUId != null && targetBUId.trim().length() > 0) {
      //WsHeaderPropagator.setTargetBUId(targetBUId, (BindingProvider)sei);
      try {
        ((SEIContext)sei).setTargetBUId(targetBUId);
      }
      catch (RuntimeException e) {
        //A nested IllegalStateException could be thrown
        //Occurs in bridge example where work is done by using UriSEILocator
        writer.println("<b>Got an RuntimeException while setting targetBUId. See log<br/>");
        logger.info("Got an RuntimeException while setting targetBUId", e);
      }
    }

    // The following method invocations are only used if registered handler
    // is not able to determine the correct value or if another value should
    // be used.
    if (interfaceGroup != null && interfaceGroup.trim().length() > 0) {
      //WsHeaderPropagator or SEIContext to set interfaceGroup for WsHeader propagation.
      //WsHeaderPropagator
      //    .setInterfaceGroup(interfaceGroup, (BindingProvider)sei);
      try {
        ((SEIContext)sei).setInterfaceGroup(interfaceGroup);
      }
      catch (RuntimeException e) {
        //A nested IllegalStateException could be thrown
        //Occurs in bridge example where work is done by using UriSEILocator
        writer.println("<b>Got an RuntimeException while setting interfaceGroup. See log<br/>");
        logger.info("Got an RuntimeException while setting interfaceGroup", e);
      }
    }
    // WsHeaderPropagator.setInterface("myOperation",
    // (BindingProvider)customerSample.getSEI());

    long start = System.currentTimeMillis();
    try {
      if (interfaceName.startsWith("ping")) {
        // The operation parameter are mapped using the so called
        // wrapper-style.
        PingResponse pingResponse = sei.ping10(new PingRequest());
        logger.finer("ping executed: address=" + pingResponse.getHostAddress()
            + "; host=" + pingResponse.getHostName());
        writer.println("<p/><b>ping executed</b><br/>");
        writer.println("<b>hostAddress:</b> " + pingResponse.getHostAddress()
            + "<br/>");
        writer.println("<b>hostName:</b> " + pingResponse.getHostName()
            + "<br/>");

      } else if (interfaceName.startsWith("createCustomer")) {
        String customerLastName = getParameter("customerLastName", "Muster");
        String customerFirstName = getParameter("customerFirstName", "Peter");
        writer.println("<b>customerName:</b> " + customerLastName + "<br/>");
        writer.println("<b>customerFirstName:</b> " + customerFirstName
            + "<br/>");
        PCustomer1 customer = new PCustomer1();
        customer.setLastName(customerLastName);
        customer.setFirstName(customerFirstName);
        customer.setSalutation("hi");
        CreateCustomer20 createcustomer20 = new CreateCustomer20();
        createcustomer20.setCustomer(customer);
        CreateCustomer20Response createcustomer20Response = sei
            .createCustomer20(createcustomer20);
        String msg = "createCustomer executed: customerId="
            + createcustomer20Response.getCustomerId();
        if (createcustomer20Response.getAttention() != null) {
          msg = msg + "; attention.exCd="
              + createcustomer20Response.getAttention().getExCd();
        }
        if (createcustomer20Response.getBusEx() != null) {
          msg = msg + "; busEx.exCd="
              + createcustomer20Response.getBusEx().getExCd();
        }
        logger.finer(msg);
        writer.println("<p/><b>" + msg + "</b>");

      } else {
        writer.println("<p/><b>interface " + interfaceName
            + " not supported!</b>");
      }

    } catch (ServiceExceptionFaultMessage sefm) {
      logger.error("got a ServiceExceptionFaultMessage", sefm);
      writer.println("<p/><div style=\"color:red;\">");
      writer.println("<b>Got a ServiceExceptionFaultMessage:</b></br><pre>");
      printServiceExceptionFaultMessage(writer, sefm);
      writer.println("</pre></div>");
    } catch (SOAPFaultException e) {
      logger.warning("got a SOAPFaultException", e);
      writer.println("<p/><div style=\"color:red;\">");
      writer.println("<b>Got a SOAPFaultException:</b></br><pre>");
      printSOAPFaultException(writer, e);
      writer.println("</pre></div>");
    } finally {
      long duration = System.currentTimeMillis() - start;
      logger.finer("Request duration (ms) " + duration);
      writer.println("</p>Time in milliseconds: " + duration + "<br/>");
    }
  }

  private static final void printSOAPFaultException(PrintWriter writer,
      SOAPFaultException e) {
    if (e.getFault() != null) {
      writer.println("FaultString=" + e.getFault().getFaultString());
      writer.println("FaultCode=" + e.getFault().getFaultCode());
      writer.println("FaultActor=" + e.getFault().getFaultActor());
      writer.println("Detail=" + e.getFault().getDetail());
    }
    writer.println("");
    e.printStackTrace(writer);
  }

  private static final void printServiceExceptionFaultMessage(
      PrintWriter writer, ServiceExceptionFaultMessage sefm) {
    if (sefm.getFaultInfo() != null) {
      writer.println("exCd=" + sefm.getFaultInfo().getExCd());
      writer.println("posInSource=" + sefm.getFaultInfo().getPosInSource());
      writer.println("sourceId=" + sefm.getFaultInfo().getSourceId());
      writer.println("timestamp=" + sefm.getFaultInfo().getTimestamp());
      if (sefm.getFaultInfo().getExParamList() != null) {
        writer.println("exParamList="
            + sefm.getFaultInfo().getExParamList().getExParam());
      }
      if (sefm.getFaultInfo().getExTraceEntryList() != null) {
        for (ExTraceEntry1 entry : sefm.getFaultInfo().getExTraceEntryList()
            .getExTraceEntry()) {
          writer.println("exTraceEntry.exCd=" + entry.getExCd());
          writer.println("exTraceEntry.sourceId=" + entry.getSourceId());
          printLogEntry1(writer, entry.getLogEntry(), "exTraceEntry.logEntry.");
        }
        printLogEntry1(writer, sefm.getFaultInfo().getLogEntry(), "logEntry.");
      }
      writer.println("");
      sefm.printStackTrace(writer);
    }
  }

  private static final void printLogEntry1(PrintWriter writer, LogEntry1 le,
      String prefix) {
    if (le != null) {
      writer.println(prefix + "logEntryId=" + le.getLogEntryId());
      writer.println(prefix + "logId=" + le.getLogId());
    }
  }

}
