package com.csg.cs.services.samples.consumer.jse;

import java.io.PrintWriter;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import com.csg.cs.core.base.common.uuid.UuidGenerator;
import com.csg.cs.core.base.config.ConfigException;
import com.csg.cs.core.base.config.ConfigHelper;
import com.csg.cs.core.base.ctx.ContextAction;
import com.csg.cs.core.base.ctx.ContextFactory;
import com.csg.cs.core.base.ctx.ContextHelper;
import com.csg.cs.core.base.ctx.CoreContext;
import com.csg.cs.core.base.ctx.CoreContextInitializer;
import com.csg.cs.core.base.ctx.RequestID;
import com.csg.cs.services.samples.consumer.CustomerSampleRequestHandler;
import com.csg.cs.services.tech_customersample_1.TECHCustomerSample1;
import com.csg.cs.services.tech_customersample_1.TECHCustomerSample1PortType;
import com.csg.cs.services.tech_customersample_2.TECHCustomerSample2;
import com.csg.cs.services.tech_customersample_2.TECHCustomerSample2PortType;
import com.csg.cs.ws.locator.SEILocator;
import com.csg.cs.ws.locator.jse.JseSEILocator;
import com.csg.cs.ws.locator.jse.WsSecurityConfigurator;

/**
 * Class that implements code suitable to run in a JSE environment (ex. JAP
 * Batch) independent of the kind of authentication.
 * <p>
 * The parameters will be read from the system properties. The list of supported
 * properties is:
 * <ul>
 * <li>customerSample1Endpoint: end-point URL of TECH_CustomerSample_1 web
 * service</li>
 * <li>customerSample2Endpoint: end-point URL of TECH_CustomerSample_2 web
 * service</li>
 * <li>See {@link CustomerSampleRequestHandler} to get a list of further
 * supported properties</li>
 * </ul>
 * 
 * @author A530732
 */
public class CustomerSampleRunner {

  /**
   * Default host and port of producer to test.
   */
  public static final String DEFAULT_HOST_PORT = "localhost:17410";

  /**
   * Default URL of end-point for TECH_CustomerSample_1.
   */
  private final static String CUSTOMER_SAMPLE_1_ENDPOINT = "https://"
      + DEFAULT_HOST_PORT + "/services/TECH_CustomerSample_1";

  /**
   * Default URL of end-point for TECH_CustomerSample_2.
   */
  private final static String CUSTOMER_SAMPLE_2_ENDPOINT = "https://"
      + DEFAULT_HOST_PORT + "/services/TECH_CustomerSample_2";

  // UUID used to generate a requestID used in CFW CoreContext.
  private static UuidGenerator uuidGen = new UuidGenerator();

  /**
   * Parameters used to handle request.
   */
  private Map<String, String[]> parameters = initParameters();

  /**
   * URL of end-point for TECH_CustomerSample_1.
   */
  private String customerSample1Endpoint = getParameter(
      "customerSample1Endpoint", CUSTOMER_SAMPLE_1_ENDPOINT);

  /**
   * URL of end-point for TECH_CustomerSample_2.
   */
  private String customerSample2Endpoint = getParameter(
      "customerSample2Endpoint", CUSTOMER_SAMPLE_2_ENDPOINT);

  /**
   * {@link SEILocator} used to lookup a TECHCustomerSample1PortType.
   */
  private JseSEILocator<TECHCustomerSample1, TECHCustomerSample1PortType> sample1Locator = null;

  /**
   * {@link SEILocator} used to lookup a TECHCustomerSample2PortType.
   */
  private JseSEILocator<TECHCustomerSample2, TECHCustomerSample2PortType> sample2Locator = null;

  private boolean AreLocatorsInitialized = false;

  /**
   * Returns {@link SEILocator} to lookup TECHCustomerSample1PortType. A
   * {@link JseSEILocator} allows to access {@link WsSecurityConfigurator}.
   */
  public final JseSEILocator<TECHCustomerSample1, TECHCustomerSample1PortType> getSample1Locator() {
    initLocators();
    return sample1Locator;
  }

  /**
   * Returns {@link SEILocator} to lookup TECHCustomerSample2PortType. A
   * {@link JseSEILocator} allows to access {@link WsSecurityConfigurator}.
   */
  public final JseSEILocator<TECHCustomerSample2, TECHCustomerSample2PortType> getSample2Locator() {
    initLocators();
    return sample2Locator;
  }

  /**
   * Initializes the used {@link SEILocator}s. Hint: Because
   * {@link JseSEILocator} optionally depends on CS_PKI the setup of the
   * {@link SEILocator} has to be done in the org.junit.Before method. The setup
   * of CS_PKI is done within {@link AbstractComponentTestCase} and is not
   * available in earlier phases.
   */
  private void initLocators() {
    if (!AreLocatorsInitialized) {
      sample1Locator = new JseSEILocator<TECHCustomerSample1, TECHCustomerSample1PortType>(
          TECHCustomerSample1.class, TECHCustomerSample1PortType.class,
          "TECH_CustomerSample_1_Port", customerSample1Endpoint);

      sample2Locator = new JseSEILocator<TECHCustomerSample2, TECHCustomerSample2PortType>(
          TECHCustomerSample2.class, TECHCustomerSample2PortType.class,
          "TECH_CustomerSample_2_Port", customerSample1Endpoint);
    }
    AreLocatorsInitialized = true;
  }

  /**
   * Just a check to ensure the JAP-CFW config-manger is configured well.
   * 
   * @throws ConfigException
   */
  @SuppressWarnings("deprecation")
  private void checkConfigMgr() throws ConfigException {
    ConfigHelper.getConfigManager().getApplicationConfig().getArea(
        "corex.ctxtransmitter.ws.consumer");
  }

  /**
   * Invokes the web service.
   */
  public void run() throws Exception {
    // demonstrates the usage of CoreContext
    CoreContextInitializer initializer = ContextHelper
        .getCoreContextInitializer();
    ContextFactory contextFactory = initializer.createContextFactory();
    contextFactory.setCurrentAppID("ICTO-1234");
    contextFactory.setOrigAppID("ICTO-11111");
    contextFactory.setRequestID(new RequestID(uuidGen.generateTimeBasedUuid()
        .toString(), 1));
    CoreContext newContext = contextFactory.createContext();

    ContextAction<Object> action = new ContextAction<Object>() {
      public Object run() throws Exception {
        checkConfigMgr();
        initLocators();
        System.out
            .println(">>>WsSecurityConfigurator.toString()="
                + ((JseSEILocator<TECHCustomerSample1, TECHCustomerSample1PortType>)sample1Locator)
                    .getWsSecurityConfigurator());

        PrintWriter w = new PrintWriter(System.out);
        // Runs same code as in case of J2EE-Servlet based consumer!
        CustomerSampleRequestHandler.createInstance(w, parameters,
            sample1Locator, sample2Locator).handleRequest();
        w.flush();
        return null;
      }
    };

    ContextHelper.getCoreContextInitializer().runInContext(newContext, action);
  }

  /**
   * Gets the value of single valued parameter. Returns defaultValue if
   * parameter is not found.
   */
  public final String getParameter(String key, String defaultValue) {
    String v = null;
    String[] vArray = parameters.get(key);
    if (vArray != null && vArray.length > 0) {
      v = vArray[0];
    }
    if (v == null) {
      return defaultValue;
    }
    return v;
  }

  private final static Map<String, String[]> initParameters() {
    Map<String, String[]> params = convertProperties(System.getProperties());
    return params;
  }

  /**
   * Converts a Properties to a Map<String, String[]> type.
   */
  private final static Map<String, String[]> convertProperties(Properties p) {
    Map<String, String[]> newP = new HashMap<String, String[]>();
    for (Map.Entry<Object, Object> e : p.entrySet()) {
      newP.put((String)e.getKey(), new String[] {(String)e.getValue()});
    }
    return newP;
  }

}
